import hashlib

from pyunique.configs import ConfigAlgo


def digest_file_bytes(filename: str) -> bytes:
    block_size = 65536
    hash_object = hashlib.new(ConfigAlgo.digest)
    with open(filename, 'rb') as file_handle:
        buf = file_handle.read(block_size)
        while len(buf) > 0:
            hash_object.update(buf)
            buf = file_handle.read(block_size)
    return hash_object.digest()



import os
from abc import ABCMeta, abstractmethod
from typing import Union

import lmdb

from pyunique.configs import ConfigLMDB, ConfigAlgo


class Archive(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def count(self) -> int:
        pass

    @abstractmethod
    def get_digest(self, filename: str) -> Union[bytes, None]:
        pass

    @abstractmethod
    def add_digest(self, filename: str, digest: bytes) -> None:
        pass

    @abstractmethod
    def start_read(self) -> None:
        pass

    @abstractmethod
    def end_read(self) -> None:
        pass

    @abstractmethod
    def start_write(self) -> None:
        pass

    @abstractmethod
    def end_write(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def yield_all_items(self):
        pass

    @abstractmethod
    def delete(self, filename: str) -> None:
        pass


class ArchiveLMDB(Archive):
    def __init__(self):
        super().__init__()
        filename = os.path.expanduser("~/.config/pyunique.lmdb")
        self.env = lmdb.open(
            path=filename,
            map_size=ConfigLMDB.map_size,
            subdir=False,  # default is True, for just one file
            writemap=True,  # default is False, for performance
            # lock=True,  # this is the default
        )
        self.txn = None

    def count(self) -> int:
        return self.env.stat()["entries"]

    def start_read(self) -> None:
        self.txn = self.env.begin(
            buffers=ConfigLMDB.buffers,
        )

    def end_read(self) -> None:
        self.txn.commit()

    def start_write(self) -> None:
        # write is False by default
        self.txn = self.env.begin(
            write=True,  # default is False
            buffers=ConfigLMDB.buffers,
        )

    def end_write(self) -> None:
        self.txn.commit()

    def get_digest(self, filename: str) -> Union[bytes, None]:
        return self.txn.get(key=filename.encode(ConfigAlgo.encoding))

    def add_digest(self, filename: str, digest: bytes) -> None:
        self.txn.put(
            key=filename.encode(ConfigAlgo.encoding),
            value=digest,
        )

    def close(self) -> None:
        self.env.close()

    def yield_all_items(self):
        with self.txn.cursor() as curs:
            for key, value in curs:
                yield key.decode(ConfigAlgo.encoding), value

    def delete(self, filename: str) -> None:
        self.txn.delete(key=filename.encode(ConfigAlgo.encoding))


def get_archive() -> Archive:
    """
    return correct archive using the currently configured algorithm
    (factory method)
    """
    return ArchiveLMDB()



"""
All configurations for pyunique
"""
from pytconf import Config, ParamCreator
import hashlib


class ConfigScan(Config):
    """
    Parameters that control the scanning phase
    """
    folder = ParamCreator.create_existing_folder(
        help_string="Which folder to work on?",
        default=".",
    )


class ConfigAlgo(Config):
    """
    Parameters at the core of the algorithm
    """
    digest = ParamCreator.create_choice(
        help_string="Which digest to use?",
        default="sha256",
        choice_list=list(hashlib.algorithms_available),
    )
    encoding = ParamCreator.create_str(
        help_string="Which encoding to use?",
        default='utf-8',
    )


class ConfigLMDB(Config):
    """
    Parameters that control working with LMDB (experts only)
    """
    map_size = ParamCreator.create_int(
        help_string="Which size of mmap to use?",
        default=1000000000000,
    )
    buffers = ParamCreator.create_bool(
        help_string="Should we use buffers? (lmdb default is False)",
        default=False,
    )



"""
The default group of operations that pyunique has
"""
import os

import tqdm
from pytconf import register_endpoint, register_function_group

import pyunique.version
from pyunique.archive import get_archive
from pyunique.configs import ConfigScan, ConfigAlgo, ConfigLMDB
from pyunique.digest import digest_file_bytes
from pyunique.utils import get_logger, get_number_of_files

GROUP_NAME_DEFAULT = "default"
GROUP_DESCRIPTION_DEFAULT = "all pyunique commands"


def register_group_default():
    """
    register the name and description of this group
    """
    register_function_group(
        function_group_name=GROUP_NAME_DEFAULT,
        function_group_description=GROUP_DESCRIPTION_DEFAULT,
    )


@register_endpoint(
    group=GROUP_NAME_DEFAULT,
)
def version() -> None:
    """
    Print version
    """
    print(pyunique.version.VERSION_STR)


@register_endpoint(
    group=GROUP_NAME_DEFAULT,
    configs=[
        ConfigScan,
        ConfigAlgo,
        ConfigLMDB,
    ],
)
def scan() -> None:
    """
    Scan a folder and register it's checksums
    """
    archive = get_archive()
    archive.start_write()
    logger = get_logger()
    logger.info("Scanning for number of files...")
    num_files = get_number_of_files(folder=ConfigScan.folder)
    with tqdm.tqdm(total=num_files) as progress_bar:
        for root, directories, files in os.walk(ConfigScan.folder):
            for file in files:
                filename = os.path.join(root, file)
                filename = os.path.abspath(filename)
                logger.debug(f"doing {filename}")
                if archive.get_digest(filename=filename) is None:
                    digest = digest_file_bytes(filename=filename)
                    archive.add_digest(filename=filename, digest=digest)
                progress_bar.update()
    archive.end_write()
    archive.close()


@register_endpoint(
    group=GROUP_NAME_DEFAULT,
    configs=[
        ConfigScan,
        ConfigAlgo,
        ConfigLMDB,
    ],
)
def clean_db() -> None:
    """
    Scan the DB and remove any entries which are no longer valid
    """
    logger = get_logger()
    archive = get_archive()
    archive.start_write()
    count = archive.count()
    errors = 0
    with tqdm.tqdm(total=count) as progress_bar:
        for filename, _digest in archive.yield_all_items():
            if not os.path.isfile(filename):
                errors += 1
                logger.debug(f"problem with filename {filename}")
                archive.delete(filename=filename)
            progress_bar.update()
    archive.end_write()
    archive.close()
    logger.info(f"number of errors is {errors}")


@register_endpoint(
    group=GROUP_NAME_DEFAULT,
    configs=[
        ConfigScan,
        ConfigAlgo,
        ConfigLMDB,
    ],
)
def check_filenames() -> None:
    """
    Check filenames in a certain folder
    """
    logger = get_logger()
    errors = 0
    error_filenames = []
    logger.info("Scanning for number of files...")
    num_files = get_number_of_files(folder=ConfigScan.folder)
    with tqdm.tqdm(total=num_files) as progress_bar:
        for root, directories, files in os.walk(ConfigScan.folder):
            for file in files:
                filename = os.path.join(root, file)
                filename = os.path.abspath(filename)
                logger.debug(f"doing {filename}")
                try:
                    filename.encode(ConfigAlgo.encoding)
                except UnicodeEncodeError:
                    errors += 1
                    error_filenames.append(filename)
                progress_bar.update()
    logger.info(f"number of errors is {errors}")
    for error_filename in error_filenames:
        logger.info(error_filename)



import logging
import os

from pyunique import LOGGER_NAME


def get_logger():
    return logging.getLogger(LOGGER_NAME)


def get_number_of_files(folder: str) -> int:
    count = 0
    for root, directories, files in os.walk(folder):
        count += len(files)
    return count
